import sys
import os

from etl_server.pipelines.models import Models as PipelineModels
from operators.dgp_kind import operator

etl_models = PipelineModels(os.environ['ETLS_DATABASE_URL'])

if __name__ == '__main__':
    pipeline_id = sys.argv[1]
    print('running dag', pipeline_id)

    for pipeline in etl_models.all_pipelines():
        dag_id = pipeline['id']
        if dag_id != pipeline_id:
            continue
        name = pipeline['name']
        params = pipeline['params']

        operator(name, params)
        break



import calendar
from urllib.parse import unquote_plus
import boto3
from botocore.exceptions import ClientError
import json
import tempfile
import shutil
from dgp_oauth2.models import get_user

class Controllers():

    def __init__(self, bucket_name, endpoint_url, aws_access_key_id, aws_secret_access_key, aws_region):
        kw = dict(
            endpoint_url=endpoint_url,
            aws_access_key_id=aws_access_key_id,
            aws_secret_access_key=aws_secret_access_key,
            region_name=aws_region
        )
        self.s3client = boto3.client('s3', **kw)
        self.s3 = boto3.resource('s3', **kw)
        self.bucket = self.s3.Bucket(bucket_name)
        exists = False
        try:
            exists = self.bucket.creation_date is not None
        except:
            pass
        if not exists:
            try:
                self.bucket.create(ACL='private')
            except:
                # Avoid race conditions
                pass

    @staticmethod
    def ownerid(obj):
        return obj.metadata.get('Ownerid') or obj.metadata.get('ownerid')

    @staticmethod
    def ownername(obj):
        return obj.metadata.get('Ownername') or obj.metadata.get('ownername')

    def list_files(self, user=None):
        filenames = [
            unquote_plus(o.key)
            for o in self.bucket.objects.all() or []
        ]
        objects = [
            self.bucket.Object(f) for f in filenames
        ]
        ret = [
            (
                o.key,
                o.last_modified,
                o.content_length,
                self.ownerid(o),
                self.ownername(o),
            )
            for o in objects
        ]
        ret = [
            dict(
                filename=filename,
                last_modified=calendar.timegm(last_modified.timetuple()),
                size=size,
                owner_id=ownerid,
                owner_name=ownername,
            )
            for filename, last_modified, size, ownerid, ownername in ret
            if user in (None, ownerid)
        ]
        return dict(
            success=True, result=ret
        )

    def download_file(self, filename):
        return self.s3client.generate_presigned_url(
            'get_object',
            dict(Bucket=self.bucket.name, Key=filename),
            HttpMethod='GET'
        )

    def upload_file(self, file_obj, filename, user, admin=False):
        profile = get_user(user)
        username = profile.get('name')
        o = self.bucket.Object(filename)
        allowed = False
        try:
            if admin or self.ownerid(o) == user:
                allowed = True
        except ClientError:
            allowed = True
        if allowed:
            metadata = dict(
                ownerid=user,
                ownername=username
            )
            with tempfile.NamedTemporaryFile() as t:
                shutil.copyfileobj(file_obj.stream, t)
                t.flush()
                t.seek(0)
                self.s3client.put_object(
                    Body=t,
                    Bucket=self.bucket.name,
                    Key=filename,
                    ContentType=file_obj.mimetype or 'application/octet-stream',
                    Metadata=metadata,
                    ACL='private'
                )
            return dict(
                success=True,
            )
        else:
            return dict(
                success=False
            )

    def delete_file(self, filename, user, admin=False):
        o = self.bucket.Object(filename)
        try:
            if admin or self.ownerid(o) == user:
                o.delete()
                return dict(
                    success=True,
                )

        except:
            pass
        return dict(
            success=False
        )



from flask import Blueprint, request, g

from .controllers import Controllers

import dgp_oauth2.credentials as credentials

from ..permissions import check_permission, Permissions


def make_blueprint(db_connection_string=None):
    """Create blueprint.
    """

    controllers = Controllers(db_connection_string)
    credentials.db_connection_string = db_connection_string
    
    # Create instance
    blueprint = Blueprint('users', 'users')

    # Controller Proxies
    @check_permission([Permissions.usersList])
    def query_users_():
        return controllers.query_users(g.permissions.get('userid'))

    @check_permission([Permissions.usersEdit, Permissions.usersNew])
    def edit_user_(role=None):
        body = request.json
        id = body['id']
        return controllers.create_or_edit_user(id, body)

    @check_permission([Permissions.usersDelete])
    def delete_user_(id):
        return controllers.delete_user(id)

    # Register routes
    blueprint.add_url_rule(
        'users', 'query_users', query_users_, methods=['GET'])
    blueprint.add_url_rule(
        'user', 'edit_user', edit_user_, methods=['POST'])
    blueprint.add_url_rule(
        'user/<id>', 'delete_user', delete_user_, methods=['DELETE'])

    # Return blueprint
    return blueprint



"""
Code that goes along with the Airflow tutorial located at:
https://github.com/apache/airflow/blob/master/airflow/example_dags/tutorial.py
"""
import sys
import os
import datetime 
import logging
import importlib

from inspect import signature

from airflow import DAG
from airflow.operators.python_operator import PythonOperator
from airflow.operators.bash_operator import BashOperator
from airflow.utils import dates
from airflow.utils.dates import cron_presets

from etl_server.pipelines.cache import Cache

def wrapper(operator, id):
    _id = id
    _operator = operator
    def func(*args):
        result = None
        try:
            print('CALLING OPERATOR')
            args = args[:len(list(signature(_operator).parameters))]
            result = _operator(*args)
            print('OPERATOR DONE, RESULT={}'.format(result))
        finally:
            from etl_server.pipelines.models import Models
            from etl_server.pipelines.controllers import Controllers
            models = Models(os.environ['ETLS_DATABASE_URL'])
            pipeline = models.query_one(_id)['result']['value']
            if result and not pipeline.get('result'):
                Controllers.trigger_event('accepted', pipeline)
            pipeline['result'] = result
            models.create_or_edit(_id, pipeline, allowed_all=True)
            print('SAVED PIPELINE={}, RESULT={}'.format(_id, result))

    return func


default_args = {
    'owner': 'Airflow',
    'depends_on_past': False,
    'start_date': dates.days_ago(1),
    'is_paused_upon_creation': False
}

for pipeline in Cache.cached_pipelines():
    dag_id = pipeline['id']
    logging.info('Initializing DAG %s', dag_id)

    schedule = pipeline['schedule']
    if isinstance(schedule, str):
        if schedule not in cron_presets:
            if len(schedule.split()) not in (5, 6):
                if '@' + schedule in cron_presets:
                    schedule = '@' + schedule
                else:
                    schedule = None
    else:
        schedule = None

    try:
        dag = DAG(dag_id, default_args=default_args, schedule_interval=schedule)

        kind = pipeline['kind']
        operator = importlib.import_module('.' + kind, package='operators').operator

        t1 = PythonOperator(task_id=dag_id,
                            python_callable=wrapper(operator, dag_id),
                            op_args=[
                                pipeline['name'],pipeline['params'], pipeline
                            ],
                            dag=dag)
        globals()[dag_id] = dag
    except Exception as e:
        logging.error(f'Failed to create a DAG with id {dag_id}, schedule {schedule} because {e}')


task_id = '_clean_scheduler_logs' 
dag_id = task_id + '_dag'
schedule = '0 * * * *'
clean_scheduler_logs_dag = DAG(dag_id, default_args=default_args,
                                schedule_interval=schedule)
clean_scheduler_logs = BashOperator(task_id=task_id,
                                    bash_command='cd /app/airflow/logs/scheduler/ && rm -rf * && echo cleaned',
                                    dag=clean_scheduler_logs_dag)

def event_proxy(handler, **kwargs):
    pipeline = kwargs['dag_run'].conf
    return handler(pipeline)

for event in ('delete', 'failed', 'new', 'submitted', 'accepted'):
    handler = importlib.import_module(f'events.{event}_pipeline').handler
    task_id = f'event_handler_{event}_pipeline'
    dag_id = task_id + '_dag'
    dag = DAG(dag_id, default_args=default_args, schedule_interval=None)
    task = PythonOperator(task_id=task_id, python_callable=event_proxy,
                          op_args=[handler], dag=dag, provide_context=True)
    globals()[dag_id] = dag



from flask import Blueprint, request

from ..permissions import check_permission, Permissions
from .controllers import Controllers


def make_blueprint(db_connection_string=None, configuration={}):  # noqa
    """Create blueprint.
    """

    controllers = Controllers(configuration=configuration,
                              connection_string=db_connection_string)

    # Create instance
    blueprint = Blueprint('pipelines', 'pipelines')

    @check_permission([Permissions.pipelinesListAll, Permissions.pipelinesListOwn, Permissions.pipelinesListPublic])
    def query_pipelines_(role=None, user=None):
        if role == Permissions.pipelinesListAll:
            return controllers.query_pipelines()
        elif role == Permissions.pipelinesListOwn:
            return controllers.query_pipelines(user=user)
        elif role == Permissions.pipelinesListPublic:
            return controllers.query_pipelines(public=True)

    @check_permission([Permissions.login])
    def configuration_():
        return controllers.configuration()

    @check_permission([Permissions.pipelinesEditAll, Permissions.pipelinesEditOwn])
    def edit_pipeline_(role=None, user=None):
        if request.method == 'POST':
            body = request.json
            id = body.get('id')
            return controllers.create_or_edit_pipeline(id, body, user, role == Permissions.pipelinesEditAll)
        else:
            return {}

    @check_permission([Permissions.pipelinesStatusAll, Permissions.pipelinesStatusOwn, Permissions.pipelinesStatusPublic])
    def query_pipeline_(id, role=None, user=None):
        if role == Permissions.pipelinesStatusAll:
            return controllers.query_pipeline(id)
        elif role == Permissions.pipelinesStatusOwn:
            return controllers.query_pipeline(id, user=user)
        elif role == Permissions.pipelinesStatusPublic:
            return controllers.query_pipeline(id, public=True)

    @check_permission([Permissions.pipelinesDeleteAll, Permissions.pipelinesDeleteOwn])
    def delete_pipeline_(id, role=None, user=None):
        if role == Permissions.pipelinesDeleteAll:
            return controllers.delete_pipeline(id)
        elif role == Permissions.pipelinesDeleteOwn:
            return controllers.delete_pipeline(id, user=user)


    @check_permission([Permissions.pipelinesExecute])
    def start_pipeline_(id):
        return controllers.start_pipeline(id)

    # Register routes
    blueprint.add_url_rule(
        'pipelines', 'query_pipelines', query_pipelines_, methods=['GET'])
    blueprint.add_url_rule(
        'pipeline', 'edit_pipeline', edit_pipeline_, methods=['POST'])
    blueprint.add_url_rule(
        'pipeline/<id>', 'query_pipeline', query_pipeline_, methods=['GET'])
    blueprint.add_url_rule(
        'pipeline/start/<id>', 'start_pipeline', start_pipeline_, methods=['POST'])
    blueprint.add_url_rule(
        'pipeline/<id>', 'delete_pipeline', delete_pipeline_, methods=['DELETE'])
    blueprint.add_url_rule(
        'configuration', 'configuration', configuration_, methods=['GET'])

    # Return blueprint
    return blueprint



from flask import Blueprint, request, g

from .controllers import Controllers

from ..permissions import check_permission, Permissions


def make_blueprint(db_connection_string=None):
    """Create blueprint.
    """

    controllers = Controllers(db_connection_string)
    
    # Create instance
    blueprint = Blueprint('datarecords', 'datarecords')

    # Controller Proxies
    @check_permission([Permissions.datarecordRead])
    def query_datarecords_(kind):
        return controllers.query(kind)

    @check_permission([Permissions.datarecordRead])
    def query_datarecord_(kind, id):
        return controllers.query_one(kind, id)

    @check_permission([Permissions.datarecordEdit, Permissions.datarecordNew])
    def edit_datarecord_(kind, role=None, user=None):
        body = request.json
        id = body['id']
        return controllers.create_or_edit(kind, id, body, user)

    @check_permission([Permissions.datarecordDelete])
    def delete_datarecord_(kind, id):
        return controllers.delete(kind, id)

    # Register routes
    blueprint.add_url_rule(
        'datarecords/<kind>', 'query_datarecords', query_datarecords_, methods=['GET'])
    blueprint.add_url_rule(
        'datarecord/<kind>/<id>', 'query_datarecord', query_datarecord_, methods=['GET'])
    blueprint.add_url_rule(
        'datarecord/<kind>', 'edit_datarecord', edit_datarecord_, methods=['POST'])
    blueprint.add_url_rule(
        'datarecord/<kind>/<id>', 'delete_datarecord', delete_datarecord_, methods=['DELETE'])

    # Return blueprint
    return blueprint



import os
import tempfile
import boto3

from dgp.core import BaseDataGenusProcessor, BaseAnalyzer
from dgp.config.consts import CONFIG_URL
from dgp_server.log import logger


_bucket = None
def bucket():
    global _bucket
    if _bucket is None:
        bucket_name = os.environ.get('BUCKET_NAME')
        endpoint_url = os.environ.get('S3_ENDPOINT_URL')
        aws_access_key_id = os.environ.get('AWS_ACCESS_KEY_ID')
        aws_secret_access_key = os.environ.get('AWS_SECRET_ACCESS_KEY')
        aws_region = os.environ.get('AWS_REGION')

        s3 = boto3.resource('s3',         
            endpoint_url=endpoint_url,
            aws_access_key_id=aws_access_key_id,
            aws_secret_access_key=aws_secret_access_key,
            region_name=aws_region
        )
        _bucket = s3.Bucket(bucket_name)
    return _bucket

def cache_dir():
    return os.path.join(tempfile.gettempdir(), 'file-cache')

os.makedirs(cache_dir(), exist_ok=True)


class FileLoaderAnalyzer(BaseAnalyzer):

    CONFIG_SOURCE_FILENAME = 'loader.filename'

    def download_out_filename(self):
        filename = self.config.get(self.CONFIG_SOURCE_FILENAME)
        if not filename:
            return
        logger.warning('FileLoaderAnalyzer filename=%s', filename)
        obj = bucket().Object(filename)
        out_filename = os.path.join(cache_dir(), '{}-{}'.format(obj.last_modified.isoformat(), filename))
        if not os.path.exists(out_filename):
            logger.warning('FileLoaderAnalyzer downloading')
            obj.download_file(Filename=out_filename)
        return out_filename

    def run(self):
        if self.cached_out_filename:
            current_url = self.config.get(CONFIG_URL)
            logger.warning('FileLoaderAnalyzer current_url=%s', current_url)
            if current_url != self.cached_out_filename:
                self.config.set(CONFIG_URL, self.cached_out_filename)
                self.context.reset_stream()

    def analyze(self):
        self.cached_out_filename = self.download_out_filename()
        self.run()
        return True


class FileLoaderDGP(BaseDataGenusProcessor):

    def __init__(self, config, context):
        super().__init__(config, context)
        self.steps = self.init_classes([
           FileLoaderAnalyzer,
        ])



import os

from sqlalchemy.ext.declarative import declarative_base

from sqlalchemy import Column, String

from dgp_oauth2.models import get_users

from ..db_utils import Common, ModelsBase


# ## SQL DB
Base = declarative_base()


# ## USERS
class Taxonomy(Base, Common):
    __tablename__ = 'etl_taxonomies'


class Models(ModelsBase):

    def __init__(self, connection_string=None):
        super().__init__(Base, Taxonomy, connection_string)




from flask import Blueprint, request, g

from .controllers import Controllers

from ..permissions import check_permission, Permissions


def make_blueprint(db_connection_string=None):
    """Create blueprint.
    """

    controllers = Controllers(db_connection_string)
    
    # Create instance
    blueprint = Blueprint('taxonomies', 'taxonomies')

    # Controller Proxies
    @check_permission([Permissions.taxonomyRead])
    def query_taxonomies_():
        return controllers.query()

    @check_permission([Permissions.taxonomyRead])
    def query_taxonomy_(id):
        return controllers.models.query_one(id)

    @check_permission([Permissions.taxonomyEdit, Permissions.taxonomyNew])
    def edit_taxonomy_(role=None):
        body = request.json
        id = body['id']
        return controllers.models.create_or_edit(id, body)

    @check_permission([Permissions.taxonomyDelete])
    def delete_taxonomy_(id):
        return controllers.models.delete(id)

    # Register routes
    blueprint.add_url_rule(
        'taxonomies', 'query_taxonomies', query_taxonomies_, methods=['GET'])
    blueprint.add_url_rule(
        'taxonomy/<id>', 'query_taxonomy', query_taxonomy_, methods=['GET'])
    blueprint.add_url_rule(
        'taxonomy', 'edit_taxonomy', edit_taxonomy_, methods=['POST'])
    blueprint.add_url_rule(
        'taxonomy/<id>', 'delete_taxonomy', delete_taxonomy_, methods=['DELETE'])

    # Return blueprint
    return blueprint



from dgp_oauth2.models import get_user, delete_user

from .models import Models


class Controllers():

    def __init__(self, connection_string):
        self.models = Models(connection_string=connection_string)

    def query(self):
        ret = self.models.query()
        ret['result'] = [x for x in ret['result'] if x['key'] != '_common_']
        return ret


import os

from sqlalchemy.ext.declarative import declarative_base

from sqlalchemy import Column, String

from dgp_oauth2.models import get_users

from ..db_utils import Common, ModelsBase


# ## SQL DB
Base = declarative_base()


# ## USERS
class DataRecord(Base, Common):
    __tablename__ = 'datarecords'
    created_by = Column(String(128))
    updated_by = Column(String(128))


class Models(ModelsBase):

    def __init__(self, connection_string=None):
        super().__init__(Base, DataRecord, connection_string)

def create_or_edit(self, key, value, user):
        return super().create_or_edit(
            key, value,
            create_kw=dict(created_by=user, updated_by=user),
            update_kw=dict(updated_by=user),
        )



import sys
import tempfile
import yaml
import os
import logging
import json

from sqlalchemy import create_engine

from dataflows import Flow, printer

from dgp.core import Config, Context
from dgp.genera import SimpleDGP, LoaderDGP, PostLoaderDGP, TransformDGP, EnricherDGP, PublisherDGP
from dgp.taxonomies import TaxonomyRegistry

from .fileloader import FileLoaderDGP


engine = None

def get_engine():
    global engine
    if engine is None:
        engine = create_engine(os.environ['DATASETS_DATABASE_URL'])
    return engine

def set_dots(o, k, v):
    k = k.split('.')
    while len(k) > 1:
        p = k.pop(0)
        o = o.setdefault(p, {})
    o[k[0]] = v


def operator(name, params, pipeline):
    with tempfile.NamedTemporaryFile(mode='w', encoding='utf-8') as config_file:
        params['dgpConfig'].setdefault('publish', {})['allowed'] = True
        metadata = params['dgpConfig'].setdefault('extra', {}).setdefault('metadata', {})
        metadata['title'] = name
        metadata['updated_at'] = pipeline['__updated_at']
        metadata['created_at'] = pipeline['__created_at']
        for k, v in params.items():
            if k.startswith('extra.'):
                set_dots(params['dgpConfig'], k, v)
        logging.info('\nCONFIGURATION:\n--------------\n%s', 
                     json.dumps(params['dgpConfig'], sort_keys=True, ensure_ascii=False, indent=2))
        yaml.dump(params['dgpConfig'], config_file)
        config_file.flush()
        config = Config(config_file.name)
        taxonomy_registry = TaxonomyRegistry('taxonomies/index.yaml')
        context = Context(config, taxonomy_registry)

        logging.getLogger().setLevel(logging.DEBUG)

        steps = [
            FileLoaderDGP,
            LoaderDGP,
            PostLoaderDGP,
            TransformDGP,
            EnricherDGP,
            PublisherDGP,
        ]

        dgp = SimpleDGP(
            config, context,
            steps=steps
        )

        ret = dgp.analyze()
        if not ret:
            logging.error('Errors:')
            logging.error('\n\t - '.join([str(x) for x in dgp.errors]))
            assert False

        # logging.info('\nCONF (POST ANALYSIS):\n--------------\n%s', 
        #              json.dumps(config._unflatten(), sort_keys=True, ensure_ascii=False, indent=2))

        logging.info('Creating Flow')
        flow = dgp.flow()
        flow = Flow(
            flow,
            printer(tablefmt='html')
        )
        logging.info('Running Flow')
        _, stats = flow.process()

        logging.info('Success')

        return stats


import os
import json
import datetime

from contextlib import contextmanager

from sqlalchemy import (
    Column, String, DateTime, types,
    create_engine, inspect
)
from sqlalchemy.orm.session import sessionmaker, Session


# ## Json as string Type
class JsonType(types.TypeDecorator):
    impl = types.Unicode

    def process_bind_param(self, value, dialect):
        return json.dumps(value)

    def process_result_value(self, value, dialect):
        if value:
            return json.loads(value)
        else:
            return None

    def copy(self, **kw):
        return JsonType(self.impl.length)


class Common():
    key = Column(String(128), primary_key=True)
    value = Column(JsonType)
    created_at = Column(DateTime, default=datetime.datetime.utcnow)
    updated_at = Column(DateTime, default=datetime.datetime.utcnow, onupdate=datetime.datetime.utcnow)


class ModelsBase():

    def __init__(self, base, cls, connection_string):
        connection_string = connection_string or os.environ.get('DATABASE_URL')
        assert connection_string is not None,\
            "No database defined, please set your DATABASE_URL env-var"
        self._sql_engine = create_engine(connection_string)
        self._sql_session = None
        self._cls = cls
        base.metadata.create_all(self._sql_engine)

    def __delete(self):
        if self._sql_session:
            self._sql_session

    @contextmanager
    def session_scope(self):
        """Provide a transactional scope around a series of operations."""
        if self._sql_session is None:
            self._sql_session = sessionmaker(bind=self._sql_engine)
        session: Session = self._sql_session()
        try:
            yield session
            session.commit()
        except Exception:
            session.rollback()
            raise
        finally:
            session.expunge_all()
            session.close()

    @staticmethod
    def object_as_dict(obj):
        return {c.key: getattr(obj, c.key)
                for c in inspect(obj).mapper.column_attrs}

    def create_or_edit(self, key, value, edit_allowed=None, create_kw={}, update_kw={}):
        ret = dict(created=False, success=True, result=value)
        with self.session_scope() as session:
            document = session.query(self._cls)\
                .filter(self._cls.key == key).first()
            if document is None:
                document = self._cls(key=key, value=value, **create_kw)
                ret['created'] = True
                session.add(document)
            else:
                if edit_allowed is None or edit_allowed(document):
                    document.value = value
                    for k, v in update_kw.items():
                        setattr(document, k, v)
                else:
                    ret['success'] = False
        return ret

    def query_one(self, key, case_sensitive=True):
        ret = dict(success=False)
        with self.session_scope() as session:
            if case_sensitive:
                document = session.query(self._cls)\
                    .filter(self._cls.key == key).first()
            else:
                document = session.query(self._cls)\
                    .filter(self._cls.key.ilike(key)).first()
            if document is not None:
                ret['success'] = True
                ret['result'] = self.object_as_dict(document)
        return ret

    def query(self, order_by=None):
        order_by = order_by or self._cls.updated_at.desc()
        ret = dict(result=[], success=True)
        with self.session_scope() as session:
            documents = session.query(self._cls).order_by(order_by)
            for doc in documents:
                ret['result'].append(self.object_as_dict(doc))
        return ret

    def delete(self, key, delete_allowed=None):
        ret = dict(success=False)
        with self.session_scope() as session:
            document = session.query(self._cls)\
                .filter(self._cls.key == key).first()
            if document is not None:
                if not delete_allowed or delete_allowed(document):
                    ret['success'] = True
                    session.delete(document)
        return ret



import os
import inspect
from enum import Enum
import logging

from flask import request, g, abort
from flask_jsonpify import jsonpify

from dgp_oauth2 import Verifyer
import dgp_oauth2.credentials as credentials


class Permissions():
    # Levels
    Viewer = 1
    Maintainer = 2
    Admin = 3

    # Roles
    login = 'login'
    
    pipelinesListPublic = 'pipelinesListPublic'
    pipelinesListOwn = 'pipelinesListOwn'
    pipelinesListAll = 'pipelinesListAll'

    pipelinesStatusPublic = 'pipelinesStatusPublic'
    pipelinesStatusOwn = 'pipelinesStatusOwn'
    pipelinesStatusAll = 'pipelinesStatusAll'

    pipelinesNew = 'pipelinesNew'
    pipelinesDeleteOwn = 'pipelinesDeleteOwn'
    pipelinesDeleteAll = 'pipelinesDeleteAll'

    pipelinesEditOwn = 'pipelinesEditOwn'
    pipelinesEditAll = 'pipelinesEditAll'

    workbench = 'workbench'

    pipelinesExecute = 'pipelinesExecute'

    usersList = 'usersList'
    usersNew = 'usersNew'
    usersEdit = 'usersEdit'
    usersDelete = 'usersDelete'

    filesListOwn = 'filesListOwn'
    filesListAll = 'filesListAll'
    filesDownload = 'filesDownload'
    filesUpload = 'filesUpload'
    filesUpdateOwn = 'filesUpdateOwn'
    filesUpdateAll = 'filesUpdateAll'
    filesDeleteOwn = 'filesDeleteOwn'
    filesDeleteAll = 'filesDeleteAll'

    taxonomyRead = 'taxonomyRead'
    taxonomyNew = 'taxonomyNew'
    taxonomyEdit = 'taxonomyEdit'
    taxonomyDelete = 'taxonomyDelete'

    datarecordRead = 'datarecordRead'
    datarecordNew = 'datarecordNew'
    datarecordEdit = 'datarecordEdit'
    datarecordDelete = 'datarecordDelete'
    
    # Level Roles
    ViewerRoles = { login, pipelinesListPublic, pipelinesStatusPublic }
    DataEditorRoles = { datarecordRead, datarecordEdit, datarecordNew }
    MaintainerRoles = ViewerRoles | DataEditorRoles | \
                        { pipelinesListOwn, pipelinesStatusOwn, pipelinesNew, pipelinesEditOwn, workbench,
                          filesListOwn, filesUpload, filesDownload, filesUpdateOwn, filesDeleteOwn, taxonomyRead }
    SuperMaintainerRoles = MaintainerRoles | { pipelinesDeleteOwn, filesListAll }
    AdminRoles = SuperMaintainerRoles | { pipelinesListAll, pipelinesStatusAll, pipelinesEditAll, pipelinesExecute, pipelinesDeleteAll,
                                          usersList, usersNew, usersEdit, usersDelete, filesUpdateAll, filesDeleteAll,
                                          taxonomyNew, taxonomyEdit, taxonomyDelete, datarecordDelete }

    # Level Roles Mapping
    Roles = {
        Viewer: ViewerRoles,
        Maintainer: MaintainerRoles,
        Admin: AdminRoles,
    }

__verifyer = None
def verifyer():
    global __verifyer
    if __verifyer is None:
        __verifyer = Verifyer(public_key=credentials.public_key)
    return __verifyer

def check_permission(roles):
    def decorator(func):
        def wrapper(*args, **kw):
            global __verifyer
            token = request.values.get('jwt') or request.headers.get('X-Auth')
            permissions = verifyer().extract_permissions(token)
            if not (permissions is False):
                level = permissions.get('permissions', {}).get('level', 0)
                user_roles = Permissions.Roles.get(level, [])
                for role in roles:
                    if role in user_roles:
                        g.permissions = permissions
                        fargs = inspect.getargspec(func).args
                        if 'role' in fargs:
                            kw['role'] = role
                        if 'user' in fargs:
                            kw['user'] = permissions['userid']
                        return func(*args, **kw)
            abort(403)
        return wrapper
    return decorator



from dgp_oauth2.models import get_user, delete_user

from .models import Models


class Controllers():

    def __init__(self, connection_string):
        self.models = Models(connection_string=connection_string)

    def create_or_edit_user(self, id, body):
        ret = self.models.create_or_edit(id, body)
        return ret

    def delete_user(self, id):
        ret = self.models.delete(id)
        auth_delete = delete_user(id)
        ret['success'] = ret['success'] or auth_delete
        return ret

    def query_users(self, userid):
        user = get_user(userid)
        if user is None:
            return {}

        userid = user.get('id')
        if userid is None:
            return {}

        all_users = self.models.all_users()

        for user in all_users['result']:
            user['self'] = user['key'] == userid

        return all_users



from slugify import slugify

from .models import Models


class Controllers():

    def __init__(self, connection_string):
        self.models = Models(connection_string=connection_string)

    def _key(self, kind, id):
        return '{}::{}'.format(kind, id)

    def create_or_edit(self, kind, id, value, user):
        id = slugify(id)
        key = self._key(kind, id)
        value['id'] = id
        return self.models.create_or_edit(key, value,
            create_kw=dict(created_by=user),
            update_kw=dict(updated_by=user))

    def query_one(self, kind, id):
        return self.models.query_one(self._key(kind, id))

    def query(self, kind):
        ret = self.models.query()
        ret['result'] = [
            x
            for x in ret['result']
            if x['key'].startswith(kind + '::')
        ]
        return ret

    def delete(self, kind, id):
        return self.models.delete(self._key(kind, id))



import os

from sqlalchemy.ext.declarative import declarative_base

from sqlalchemy import Column, String

from dgp_oauth2.models import get_users

from ..db_utils import Common, ModelsBase


# ## SQL DB
Base = declarative_base()


# ## USERS
class User(Base, Common):
    __tablename__ = 'etl_users'


class Models(ModelsBase):

    def __init__(self, connection_string=None):
        super().__init__(Base, User, connection_string)

    def query_one(self, key):
        return super().query_one(key, case_sensitive=False)

    def all_users(self):
        auth_users = get_users()
        auth_users = dict(
            (x['id'], x) for x in auth_users
        )
        etl_users = self.query()

        for user in etl_users['result']:
            for auth_user in auth_users.keys():
                if auth_user == user['key']:
                    user['value'].update(auth_users[auth_user])
                    auth_users.pop(auth_user, None)
                    break
        for auth_user in auth_users.values():
            etl_users['result'].append(dict(
                key=auth_user['id'], 
                value=auth_user
            ))

        return etl_users



import os
from flask import Blueprint, request, g, redirect

from .controllers import Controllers

import dgp_oauth2.credentials as credentials

from ..permissions import check_permission, Permissions


def make_blueprint(bucket_name, endpoint_url, aws_access_key_id, aws_secret_access_key, aws_region):
    """Create blueprint.
    """

    controllers = Controllers(bucket_name, endpoint_url, aws_access_key_id, aws_secret_access_key, aws_region)
    
    # Create instance
    blueprint = Blueprint('files', 'files')

    # Controller Proxies
    @check_permission([Permissions.filesListAll, Permissions.filesListOwn])
    def list_files_(role=None, user=None):
        if role == Permissions.filesListAll:
            return controllers.list_files()
        elif role == Permissions.filesListOwn:
            return controllers.list_files(user)

    @check_permission([Permissions.filesDownload])
    def download_file_():
        filename = request.values['filename']
        return redirect(controllers.download_file(filename), code=301)

    @check_permission([Permissions.filesUpdateAll, Permissions.filesUpdateOwn, Permissions.filesUpload])
    def upload_file_(role=None, user=None):
        filename = os.path.basename(request.values['filename'])
        file_obj = request

        if role == Permissions.filesUpdateAll:
            ret = controllers.upload_file(file_obj, filename, user, admin=True)
        elif role in [Permissions.filesUpdateOwn, Permissions.filesUpload]:
            ret = controllers.upload_file(file_obj, filename, user, admin=False)
        return ret

    @check_permission([Permissions.filesDeleteAll, Permissions.filesDeleteOwn])
    def delete_file_(role=None, user=None):
        filename = request.values['filename']
        if role == Permissions.filesDeleteAll:
            ret = controllers.delete_file(filename, user, admin=True)
        elif role == Permissions.filesDeleteOwn:
            ret = controllers.delete_file(filename, user, admin=False)
        return ret

    # Register routes
    blueprint.add_url_rule(
        'files', 'list_files', list_files_, methods=['GET'])
    blueprint.add_url_rule(
        'file', 'download_file', download_file_, methods=['GET'])
    blueprint.add_url_rule(
        'upload', 'upload_file', upload_file_, methods=['POST'])
    blueprint.add_url_rule(
        'file', 'delete_file', delete_file_, methods=['DELETE'])

    # Return blueprint
    return blueprint



import logging
import time
import re
import os
import json

import slugify

from .models import Models
from .cache import Cache

from airflow.models import TaskInstance, DagBag, DagRun, DagModel
from airflow.utils.db import create_session
from airflow import settings

dagbag = DagBag(settings.DAGS_FOLDER, store_serialized_dags=settings.STORE_SERIALIZED_DAGS)


TABLE_RE = re.compile('<table>.+</table>', re.MULTILINE | re.DOTALL)


class Controllers():

    def __init__(self, configuration, connection_string):
        self.models = Models(connection_string=connection_string)
        self.config = configuration
        if 'schedules' not in configuration:
            configuration['schedules'] = [
                dict(name='@hourly', display='Hourly'),
                dict(name='@daily', display='Daily'),
                dict(name='@weekly', display='Weekly'),
                dict(name='@monthly', display='Monthly'),
                dict(name='@yearly', display='Yearly'),
                dict(name='manual', display='Manual'),
            ]
        Cache.refresh_cached_pipelines(self.models)
        self.unpause_all()

    def unpause_all(self):
        dagbag = DagBag(settings.DAGS_FOLDER)
        dag_ids = dagbag.dag_ids
        for dag_id in dag_ids:
            model = DagModel.get_dagmodel(dag_id)
            if model:
                model.set_is_paused(is_paused=False)
            else:
                logging.warning('Failed to unpause dag %s', dag_id)
        

    def create_or_edit_pipeline(self, id, body, owner, allowed_all):
        # Calculate id if necessary
        if not id:
            title = body['name']
            title = title + ' ' + hex(int(time.time()))[2:]
            id = slugify.slugify(title, separator='-', lowercase=True)
        body['id'] = id

        submitted = False
        if body.get('params', {}).get('dgpConfig', {}).get('__submit'):
            submitted = True
            body['params']['dgpConfig']['__submit'] = False

        # Add record to DB
        ret = self.models.create_or_edit(id, body, owner=owner, allowed_all=allowed_all)
        Cache.refresh_cached_pipelines(self.models)

        if ret.get('created'):
            self.trigger_event('new', ret['result'])
        elif submitted:
            self.trigger_event('submitted', ret['result'])

        return ret

    def delete_pipeline(self, id, user=None):
        # Delete pipeline from DB
        pipeline = self.models.query_one(id)
        pipeline = pipeline.get('result', {}).get('value')
        ret = self.models.delete(id, user)
        Cache.refresh_cached_pipelines(self.models)
        if ret.get('success'):
            self.trigger_event('delete', pipeline)
        return ret

    def __get_latest_runs(self):
        dagruns = DagRun.get_latest_runs()
        statuses = {}
        for run in dagruns:
            statuses[run.dag_id] = dict(
                execution_date=run.execution_date and run.execution_date.isoformat(),
                start_date=run.start_date and run.start_date.isoformat(),
                end_date=run.end_date and run.end_date.isoformat(),
                status=str(run._state)
            )
        return statuses

    def __get_logs(self, id):
        logger = logging.getLogger('airflow.task')
        handler = next((handler for handler in logger.handlers
                        if handler.name == 'task'), None)
        with create_session() as session:
            ti = session.query(TaskInstance)\
                    .filter(TaskInstance.dag_id == id,
                            TaskInstance.task_id == id)\
                    .order_by(TaskInstance.execution_date.desc())\
                    .first()
        if ti is not None:
            dag = dagbag.get_dag(id)
            if dag:
                ti.task = dag.get_task(ti.task_id)
                logs, _ = handler.read(ti, None, metadata={})
            else:
                logs = ''
            if isinstance(logs, list):
                pre, post = logs[:50], logs[50:]
                post = post[-500:]
                logs = ''.join(pre + ['...'] + post)
            if not logs:
                return '', ''
            table = TABLE_RE.findall(logs)
            if len(table) > 0:
                table = table[-1]
                logs = TABLE_RE.sub('', logs)
            else:
                table = ''
            return logs, table
        else:
            return '', ''

    def query_pipelines(self, user=None, public=None):
        query_results = self.models.query()
        results = query_results.get('result', [])
        if user:
            results = list(filter(lambda p: not p['private'] or p['owner'] == user, results))
        elif public:
            results = list(filter(lambda p: not p['private'], results))
    
        statuses = self.__get_latest_runs()
        for res in results:
            status = dict(status='didnt-run')
            status = statuses.get(res['key'], status)
            res['value']['status'] = status
        query_results['result'] = list(map(lambda x: dict(x.get('value'), owner=x.get('owner'), private=x.get('private')), results))

        return query_results

    def query_pipeline(self, id, user=None, public=None):
        ret = self.models.query_one(id)
        result = ret.setdefault('result', {})
        value = result.setdefault('value', {})
        if (user and result['owner'] != user and result['private']) or (public and result['private']):
            return dict(success=False)

        statuses = self.__get_latest_runs()
        key = result.get('key')
        if key in statuses:
            status = statuses[key]
            status['logs'], status['table'] = self.__get_logs(key)
        else:
            status = dict(status='didnt-run', logs='')
        value.setdefault('status', status)
        ret['result'] = value
        return ret

    def configuration(self):
        return dict(result=self.config)

    @staticmethod
    def start_pipeline(id):
        from airflow.api.common.experimental.trigger_dag import trigger_dag
        from airflow import models
        models.DagModel.get_dagmodel(id).set_is_paused(is_paused=False)
        run = trigger_dag(id)
        return dict(result=run.run_id if run else None) 

    @staticmethod
    def trigger_event(event, pipeline):
        from airflow.api.common.experimental.trigger_dag import trigger_dag
        from airflow import models
        dag_id = f'event_handler_{event}_pipeline_dag'
        models.DagModel.get_dagmodel(dag_id).set_is_paused(is_paused=False)
        trigger_dag(dag_id, conf=pipeline)




import json

class Cache():

    CACHED_PIPELINES_FILENAME = 'cached-pipelines.json'

    @classmethod
    def refresh_cached_pipelines(cls, models):
        return json.dump(models.all_pipelines(), open(cls.CACHED_PIPELINES_FILENAME, 'w'))

    @classmethod
    def cached_pipelines(cls):
        try:
            return json.load(open(cls.CACHED_PIPELINES_FILENAME))
        except:
            return []
    


import os
from dgp_oauth2.models import get_user
from .users.models import Models
from .permissions import Permissions

default_level = int(os.environ.get('DGP_APP_DEFAULT_ROLE', 0))
default_level_ret = dict(level=default_level)

def get_permissions(service, userid):

    models = Models(os.environ['ETLS_DATABASE_URL'])

    if service != 'etl-server': return {'error':'unknown-service'}
   
    value = models.query_one(userid)
    if value is None or not value.get('success'):
        if default_level:
            ret = default_level_ret
        else:
            return {'error':'unknown-user-' + userid}
    else:
        ret = value.get('result', {}).get('value')
    if ret is None: return {'error': 'no-value'}

    if 'level' in ret:
        level = ret['level']
        ret['roles'] = list(Permissions.Roles.get(level, []))

    return ret



import os

from sqlalchemy import String, Boolean, Column
from sqlalchemy.ext.declarative import declarative_base

from ..db_utils import Common, ModelsBase, JsonType


# ## SQL DB
Base = declarative_base()


# ## Pipeline Model
class Pipeline(Base, Common):
    __tablename__ = 'etl_pipeline'
    owner = Column(String(128))
    private = Column(Boolean)


class Models(ModelsBase):

    def __init__(self, connection_string=None):
        super().__init__(Base, Pipeline, connection_string)

    def create_or_edit(self, key, value, owner=None, allowed_all=False):
        if 'status' in value:
            value.pop('status', None)
        private = value.get('private') is not False
        return super().create_or_edit(
            key, value,
            edit_allowed=lambda document: allowed_all or document.owner == owner,
            create_kw=dict(private=private, owner=owner),
            update_kw=dict(private=private),
        )

    def _pipelines(self):
        return self.query()['result']

    def all_pipelines(self):
        return list(map(
            lambda p: dict(
                **p['value'],
                __created_at=p['created_at'].isoformat() if p['created_at'] else None,
                __updated_at=p['updated_at'].isoformat() if p['updated_at'] else None,
                __owner=p['owner'],
                __private=p['private']
            ),
            self._pipelines()
        ))

    def delete(self, id, user):
        return super().delete(id, delete_allowed=lambda p: user is None or p['owner'] == user)